package jwebcache;

import org.apache.log4j.Logger;

import java.util.*;
import java.io.*;
import java.net.*;

import net.i2p.client.streaming.I2PSocket;
import net.i2p.data.Destination;


class Worker extends Thread {

    // Our logger instance.
    private static Logger logger = Logger.getLogger(Worker.class);
    
    // Configuration.
    private static Properties properties; // Our config properties.
    private static int httpReadTimeout; // Read timeout we employ.
    private static boolean logHeaders; // Should we log headers.
    private static boolean respondToQuotaExcess; // Should we bother responding to quota excess?
    
    // References to objects shared across all workers.
    private static Database database; // Database we use.
    private static WorkerPool workerPool; // Worker pool we belong to.
    private static ResponseFormatter responseFormatter; // Response formatter we use.
    
    // Other variables.
    private long requestId; // Request ID which we currently service.
    private String fromDestination; // I2P destination the request came from.
    private I2PSocket socket; // I2PSocket which we currently handle.
    
    private Timer timer; // Timer to schedule timeouts.
    private BufferedReader reader = null; // For reading requests.
    private PrintWriter writer = null; // For writing responses.
    
    // Constants signifying request types
    private static final int FAILED_OPENING_COMMS = 0;
    private static final int FAILED_READING = 1;
    
    private static final int DENIED_TOTALQUOTA = 2;
    
    private static final int FAILED_EXTRACTING = 3;
    private static final int FAILED_URLDECODING = 4;
    private static final int FAILED_CLASSIFYING = 5;
    
    private static final int DENIED_PUTQUOTA = 6;
    
    private static final int FAILED_PARSING = 7;
    private static final int FAILED_COMPOSING = 8;
    
    private static final int GET_DESTINATIONS = 9;
    private static final int GET_WEBCACHES = 10;
    private static final int PUT_DESTINATION = 11;
    private static final int PUT_WEBCACHE = 12;
    private static final int PUT_BOTH = 13;
    private static final int BROWSE_INDEX = 14;
    private static final int BROWSE_STATISTICS = 15;

    // Constants signifying response types
    private static final int RESPONSE_DROPPED = 0;
    private static final int RESPONSE_ERROR = 1;
    private static final int RESPONSE_REJECTED = 2;
    private static final int RESPONSE_ACCEPTED = 3;

    // Shutdown flag, if raised and wait() interrupted, run() will quit.
    // Should be accessed using the getQuitting() and setQuitting() methods.
    private Boolean quitting = false;


    Worker(WorkerPool workerPool, Database database, ResponseFormatter responseFormatter, Properties properties) throws Exception {
        
        // Grab arguments to local variables
        this.workerPool = workerPool;
        this.database = database;
        this.responseFormatter = responseFormatter;
        this.properties = properties;
        
        try {
            // Parse configuration properties
            httpReadTimeout = Integer.parseInt(properties.getProperty("Worker.httpReadTimeoutMs","45000"));
            logHeaders = Boolean.parseBoolean(properties.getProperty("Worker.logHeaders","false"));
            respondToQuotaExcess = Boolean.parseBoolean(properties.getProperty("Worker.respondToQuotaExcess","true"));
        } catch (Exception e) {
            logger.fatal("Failed reading config properties, worker constructor aborting.",e);
            throw new Exception("Worker constructor aborting.");
        }

        socket = null; // Initialize socket to null
        timer = new Timer(); // Initialize our timer
    }


    // Synchronized to prevent concurrence with run().
    public synchronized void assignJob(I2PSocket socket, long requestId, String fromDestination) {
        
        logger.info(this.getName() + " got assignment for request " + requestId + ".");
        
        this.socket = socket; // Accept socket assignment
        this.requestId = requestId; // Accept request reference
        this.fromDestination = fromDestination; // Accept from destination reference
        notifyAll(); // Fire notification to unblock the waiting loop
    }

    
    public boolean getQuitting() {
        synchronized(quitting) { return quitting.booleanValue(); }
    }
    
    
    public void setQuitting(boolean value) {
        synchronized(quitting) { quitting = value; }
    }


    public void quit() {

        logger.info(this.getName() + " got quit request.");
        
        logger.debug(this.getName() + " raising quit flag, interrupting wait.");
        setQuitting(true);
        interrupt();

        logger.debug(this.getName() + " quit method closing socket (if any).");
        if (socket != null) {
            try { socket.close(); }
            catch (Exception e) { logger.warn(this.getName() + " failed closing socket in quit method."); }
            finally { socket = null; }
        }
    }
    
    
    private void shutdown() {
        
        // Release timer
        timer.cancel();
        timer = null;
    }


    // Synchronized to prevent concurrence with assignJob(). Lock yielded during wait().
    public synchronized void run() {

        while (true) {
            
            // If we lack a socket, start waiting: we'll be notified.
            // If we have a socket, assignJob() has run already, so start work.
            if (socket == null) {
                try {
                    wait();
                } catch (InterruptedException e) { logger.debug(this.getName() + " interrupted."); }
            
                logger.debug(this.getName() + " unblocked.");                

                if (getQuitting()) {
                    logger.info(this.getName() + " interrupted by quit request, quitting!");
                    workerPool.acceptQuit(this);
                    shutdown();
                    return;
                }
            }
            
            // Begin work only if we absolutely have a socket.
            if (socket != null) {
            
                logger.info(this.getName() + " received socket for request " + requestId + ".");
                handleClient();
                
                // We've been marked busy, try to get recycled.
                if (!workerPool.acceptDone(this)) {
                    logger.debug(this.getName() + " couldn't get recycled, quitting.");
                    shutdown();
                    return;
                }
                else logger.debug(this.getName() + " got recycled.");
            }
        }
    }
    
    
    private void handleClient() {
        
        String request = null;
        String response = null;

        if (openClientComms()) {
            logger.info("Opened comms for request " + requestId + ", reading.");
            request = readRequest();
                
            if (request != null) {
                if (logHeaders) {
                    logger.info("Logging headers for request " + requestId + ".");
                    database.logHeaders(requestId,request);
                }

                logger.info("Composing response to request " + requestId + ".");
                response = composeResponse(request);
                    
                if (response != null) {
                    logger.info("Composed response to request " + requestId + ", sending to client.");
                    sendResponse(response);
                } else {
                    logger.error("Failed composing response to request " + requestId +  ", dropping client!");
                    database.updateRequest(requestId,FAILED_COMPOSING,RESPONSE_DROPPED);
                }
            } else {
                logger.error("Failed reading request " + requestId + ", dropping client!");
                database.updateRequest(requestId,FAILED_READING,RESPONSE_DROPPED);
            }
        } else {
            logger.error("Failed opening comms for request " + requestId + ", dropping client!");
            database.updateRequest(requestId,FAILED_OPENING_COMMS,RESPONSE_DROPPED);
        }
            
        closeClientComms();
        logger.info("Closed comms for request " + requestId + ", done handling client.");
    }
    
    
    private boolean openClientComms() {
        
        if (socket == null) {
            logger.error("Null socket cannot have streams!");
            return false;
        }
        
        try {
            logger.debug("Getting input stream reader.");
            reader = new BufferedReader(new InputStreamReader(socket.getInputStream()));
        } catch (IOException e) {
            logger.error("Failed getting input stream.",e);
            return false;
        }

        try {
            logger.debug("Getting output stream writer.");
            writer = new PrintWriter(new BufferedOutputStream(socket.getOutputStream()));
        } catch (IOException e) {
            logger.error("Failed getting output stream.",e);
            return false;
        }

        return true;
    }
    
    
    private void closeClientComms() {

        if (reader != null) {
            try {
                logger.debug("Closing input reader for " + requestId + ".");
                reader.close();
            } catch (Exception e) {
                logger.error("Failed closing input reader for " + requestId + ".",e);
            } finally {
                reader = null;
            }
        } else logger.warn("Input reader for " + requestId + " already null.");
        
        if (writer != null) {
            try {
                logger.debug("Closing output writer for " + requestId + ".");
                writer.close();
            } catch (Exception e) {
                logger.error("Failed closing output writer for " + requestId + ".",e);
            } finally {
                writer = null;
            }
        } else logger.warn("Output writer for " + requestId + " already null.");
        
        if (socket != null) {
            try {
                if (!socket.isClosed()) {
                    logger.debug("Closing socket for " + requestId + ".");
                    socket.close();
                } else logger.warn("Socket for " + requestId + " already closed.");

            } catch (IOException e) {
                logger.error("Failed closing socket for " + requestId + ".",e);
            } finally {
                socket = null;
            }
        } else logger.warn("Socket for " + requestId + " already null.");
    }
    
    
    private String readRequest() {

        SocketTimeout timeout;
        StringBuffer buffer = new StringBuffer("");
        String line;
        
        logger.info("Scheduling timeout for request " + requestId + " in " + httpReadTimeout + " ms.");
        timeout = new SocketTimeout(socket,requestId);
        timer.schedule(timeout,httpReadTimeout);
        
        do {
            // Read line, abort if exception occurs
            try {
                line = reader.readLine();
            } catch (Exception e) {
                logger.error("Failed reading line for " + requestId + ", exception caught.",e);
                timeout.cancel(); // Cancel timeout before returning
                return null;
            }

            // Check line, abort if null occurs
            if (line == null) {
                logger.info("Failed reading line for " + requestId + ", null line received.");
                timeout.cancel(); // Cancel timeout before returning
                return null;
            }

            // Add line to request, append CR+LF back onto it
            logger.debug("Received: " + line);
            buffer.append(line);
            buffer.append("\r\n");

        // Empty line indicates end of HTTP headers
        } while (!line.equals(""));

        logger.info("Request " + requestId + " received, canceling timeout.");
        timeout.cancel();
        return buffer.toString();
    }

    
    private void sendResponse(String response) {
        
        try {
            writer.print(response);
            // flushing causes slowness and has proven needless
        } catch (Exception e) {
            logger.error("Failed sending response for " + requestId + ".",e);
        }
    }
    
  
    private int classifyRequest(String request) {

        // Must start with "GET " and contain "HTTP/"
        if ((request.indexOf("GET ") != 0) || (request.indexOf(" HTTP/") < 0))
         return FAILED_CLASSIFYING;
        
        if ((request.startsWith("GET / HTTP/")) ||
            (request.startsWith("GET /introduction.html HTTP/"))) return BROWSE_INDEX;

        if (request.startsWith("GET /statistics.html HTTP/")) return BROWSE_STATISTICS;
        
        if ((request.indexOf("ip=") >= 0) &&
            (request.indexOf("url=") >= 0)) return PUT_BOTH;
        
        if (request.indexOf("ip=") >= 0) return PUT_DESTINATION;
        if (request.indexOf("url=") >= 0) return PUT_WEBCACHE;
        
        if (request.indexOf("hostfile=1") >= 0) return GET_DESTINATIONS;
        if (request.indexOf("urlfile=1") >= 0) return GET_WEBCACHES;
        
        return FAILED_CLASSIFYING;
    }
    
    
    private String composeResponse(String request) {

        // Check for total quota excess.
        if (database.exceedsTotalQuota(fromDestination)) return deniedTotalQuota();

        // Proceed only if extracting the first line succeeds.
        if ((request = extractLine(request)) == null) return failedExtracting();
        
        // Proceed only if URL decoding the first line succeeds.
        if ((request = URLDecode(request)) == null) return failedDecoding();

        // Classify the request.
        int type = classifyRequest(request);
        logger.debug("Request " + requestId + " classified as " + type + ".");
        
        // Now it's possible to enforce PUT quotas.
        if ((type == PUT_DESTINATION) || (type == PUT_WEBCACHE) || (type == PUT_BOTH))
         if (database.exceedsPutQuota(fromDestination)) deniedPutQuota();
        
        // Proceed according to classification.
        switch (type) {
            case BROWSE_INDEX:      return browseIndex();
            case BROWSE_STATISTICS: return browseStatistics();
            case GET_DESTINATIONS:  return getDestinations();
            case GET_WEBCACHES:     return getWebcaches();
            case PUT_DESTINATION:   return putDestination(request);
            case PUT_WEBCACHE:      return putWebcache(request);
            case PUT_BOTH:          return putBoth(request);
            default:                return failedClassifying();
        }
    }
    

    private String failedExtracting() {
        logger.warn("Failed extracting first line from " + requestId + ".");
        database.updateRequest(requestId,FAILED_EXTRACTING,RESPONSE_ERROR);
        return responseFormatter.formatError("Failed extracting the first line.");
    }
    
    
    private String failedDecoding() {
        logger.warn("Failed URL decoding first line.");
        database.updateRequest(requestId,FAILED_URLDECODING,RESPONSE_ERROR);
        return responseFormatter.formatError("Failed URL decoding the first line.");
    }

    
    private String failedClassifying() {
        logger.warn("Failed classifying request " + requestId + ".");
        database.updateRequest(requestId,FAILED_CLASSIFYING,RESPONSE_ERROR);
        return responseFormatter.formatError("Failed classifying the request.");
    }

    
    private String deniedTotalQuota() {
        if (respondToQuotaExcess) {
            logger.info("Request " + requestId + " exceeds total quota, rejected.");
            database.updateRequest(requestId,DENIED_TOTALQUOTA,RESPONSE_REJECTED);
            return responseFormatter.formatError("Request exceeds total quota, rejected.");
        } else {
            logger.info("Request " + requestId + " exceeds total quota, dropped.");
            database.updateRequest(requestId,DENIED_TOTALQUOTA,RESPONSE_DROPPED);
            return ""; // Returning null would get interpreted as failure
        }
    }
    
    
    private String deniedPutQuota() {
        if (respondToQuotaExcess) {
            logger.info("Request " + requestId + " exceeds PUT quota, rejected.");
            database.updateRequest(requestId,DENIED_PUTQUOTA,RESPONSE_REJECTED);
            return responseFormatter.formatError("Request exceeds PUT quota, rejected.");
        } else {
            logger.info("Request " + requestId + " exceeds PUT quota, dropped.");
            database.updateRequest(requestId,DENIED_PUTQUOTA,RESPONSE_DROPPED);
            return ""; // Returning null would get interpreted as failure
        }
    }

    
    private String browseIndex() {
        logger.info("Request " + requestId + " looks like a BROWSE INDEX request.");
        database.updateRequest(requestId,BROWSE_INDEX,RESPONSE_ACCEPTED);
        return responseFormatter.formatIntroduction();
    }

    
    private String browseStatistics() {
        logger.info("Request " + requestId + " looks like a BROWSE STATISTICS request.");
        database.updateRequest(requestId,BROWSE_STATISTICS,RESPONSE_ACCEPTED);
        return responseFormatter.formatStatistics();
    }

    
    private String getDestinations() {
        logger.info("Request " + requestId + " looks like a GET DESTINATIONS request.");
        database.updateRequest(requestId,GET_DESTINATIONS,RESPONSE_ACCEPTED);
        return responseFormatter.addHeader(database.getDestinations(),"200 OK","text/plain");
    }

    
    private String getWebcaches() {
        logger.info("Request " + requestId + " looks like a GET WEBCACHES request.");
        database.updateRequest(requestId,GET_WEBCACHES,RESPONSE_ACCEPTED);
        return responseFormatter.addHeader(database.getWebcaches(),"200 OK","text/plain");
    }

    
    private String putDestination(String request) {
        
        logger.info("Parsing PUT DESTINATION request " + requestId + ".");
        Destination dest = extractDestination(request);
        
        if (dest == null) {
            logger.warn("PUT DESTINATION request " + requestId + " failed parsing.");
            database.updateRequest(requestId,FAILED_PARSING,RESPONSE_ERROR);
            return responseFormatter.formatError("PUT DESTINATION request failed parsing.");
        } else {
            logger.info("PUT DESTINATION request " + requestId + " parsed OK.");
            database.updateRequest(requestId,PUT_DESTINATION,RESPONSE_ACCEPTED);
            database.putDestination(requestId,dest.toBase64());
            return responseFormatter.addHeader("OK: PUT DESTINATION request accepted.","200 OK","text/plain");
        }
    }

    
    private String putWebcache(String request) {

        logger.info("Parsing PUT WEBCACHE request " + requestId + ".");
        String url = extractURL(request);
        
        if (url == null) {
            logger.warn("PUT WEBCACHE request " + requestId + " failed parsing.");
            database.updateRequest(requestId,FAILED_PARSING,RESPONSE_ERROR);
            return responseFormatter.formatError("PUT WEBCACHE request failed parsing.");
        } else {
            logger.info("PUT WEBCACHE request " + requestId + " parsed OK.");
            database.updateRequest(requestId,PUT_WEBCACHE,RESPONSE_ACCEPTED);
            database.putWebcache(requestId,url);
            return responseFormatter.addHeader("OK: PUT WEBCACHE request accepted.","200 OK","text/plain");
        }
    }
    
    
    private String putBoth(String request) {

        logger.info("Parsing PUT BOTH request " + requestId + ".");
        Destination dest = extractDestination(request);
        String url = extractURL(request);
        
        if ((dest == null) || (url == null)) {
            logger.warn("PUT BOTH request " + requestId + " failed parsing.");
            database.updateRequest(requestId,FAILED_PARSING,RESPONSE_ERROR);
            return responseFormatter.formatError("PUT BOTH request failed parsing.");
        } else {
            logger.info("PUT BOTH request " + requestId + " parsed OK.");
            database.updateRequest(requestId,PUT_BOTH,RESPONSE_ACCEPTED);
            database.putDestination(requestId,dest.toBase64());
            database.putWebcache(requestId,url);
            return responseFormatter.addHeader("OK: PUT BOTH request accepted.","200 OK","text/plain");
        }
    }

    
    private String extractLine(String data) {
        if ((data.indexOf("\r\n")) > 0)
         return data.substring(0,data.indexOf("\r\n")); else
         return null;
    }

    
    private Destination extractDestination(String data) {
        
        final int DESTINATION_LENGTH = 516;

        // Destination must start with an "ip=" string.
        int start = data.indexOf("ip=") + "ip=".length();
        int expectedEnd = start + DESTINATION_LENGTH;

        if (data.length() < expectedEnd) {
            logger.warn("Not enough data to read destination!");
            return null;
        }

        // I2P destination object must form successfully.
        try {
            return new Destination(data.substring(start,expectedEnd));
        } catch (Exception e) {
            logger.warn("Failed reading destination!",e);
            return null;
        }
    }
    
    
    private String extractURL(String data) {
        
        // URL must start with an "url=" string.
        int start = data.indexOf("url=") + "url=".length();
        
        // URL must end by the next parameter, or by the end all parameters.
        int end = data.indexOf("&",start);
        if (end < 0) end = data.indexOf(" HTTP/",start);
        if (end < 0) return null;
        
        return (data.substring(start,end));
    }
    
    
    private String URLDecode(String url) {

        try {
            URLDecoder decoder = new URLDecoder();
            return decoder.decode(url,"ISO-8859-1");
        } catch (Exception e) {
            logger.error("Failed decoding URL: " + url,e);
            return null;
        }
    }
    
    
}

