package jwebcache;

import org.apache.log4j.Logger;

import java.util.Vector;
import java.util.Properties;
import java.util.Date;
import java.text.SimpleDateFormat;


import net.i2p.client.streaming.I2PSocket;


public class WorkerPool {
    
    // Our logger instance.
    private Logger logger = Logger.getLogger(WorkerPool.class);
    
    // Configuration.
    private Properties properties; // Our config properties.
    private Integer targetWorkerCount; // Minimum number of workers to keep around.
    
    // NOTE: All synchronization in WorkerPool occurs on "targetWorkerCount".
    // NOTE: That variable is essentially our mutex.
    
    // References to important objects.
    private Database database; // Database object we relay to workers.
    private ResponseFormatter responseFormatter; // Response formatter object we relay to workers.
    
    // Other variables.
    private Vector<Worker> idleWorkers = new Vector<Worker>(); // List of idle workers.
    private Vector<Worker> busyWorkers = new Vector<Worker>(); // List of busy workers.
    private boolean shuttingDown = false; // Worker pool shutting down and refusing jobs?
    private long workerId; // Unique ID for worker thread, so logs would be useful.

   
    public WorkerPool(Database database, ResponseFormatter responseFormatter, Properties properties) throws Exception {
        
        // Grab arguments to local variables
        this.database = database;
        this.responseFormatter = responseFormatter;
        this.properties = properties;
        
        try {
            // Parse configuration properties
            targetWorkerCount = Integer.parseInt(properties.getProperty("WorkerPool.targetWorkerCount","5"));
        } catch (Exception e) {
            logger.fatal("Failed reading config properties, worker pool constructor aborting.",e);
            throw new Exception("Worker pool constructor aborting.");
        }
        
        try {
            // Iterate pre-creating workers
            for (workerId = 0; workerId < targetWorkerCount; workerId++) {

                // Create worker, hand over references to necessary tools
                // Start worker, and register as idle
                logger.info("Creating worker thread " + workerId + ".");
                Worker w = new Worker(this,database,responseFormatter,properties);
                w.setName("Worker " + workerId);
                w.start();
                idleWorkers.addElement(w);
            }
        } catch (Exception e) {
            logger.fatal("Failed creating worker, worker pool constructor aborting.",e);
            shutdown();
            throw new Exception("Worker pool constructor aborting.");
        }
    }
    
    
    public void shutdown() {
        
        final int DECREMENT = 5000; // How long we sleep at a time.
        int remaining; // How much longer we may wait for workers to quit.
                        
        synchronized (targetWorkerCount) {
            // It must be guaranteed that nobody can assign requests to workers anymore.
            // If server sockets are properly closed before calling us,
            // this shouldn't happen, but better safe than sorry.
            shuttingDown = true;
            
            // Finished workers should now quit instead of getting recycled.
            targetWorkerCount = 0;
        }

        // Busy workers may quit slowly, give them 45 seconds
        remaining = 45000;
        while (remaining > 0) {
            
            synchronized (targetWorkerCount) {
                if (busyWorkers.size() == 0) break;
                logger.info("Sending quit requests to " + busyWorkers.size() + " busy workers.");
                for (int i=0; i < busyWorkers.size(); i++) busyWorkers.elementAt(i).quit();
            }
            
            // Spend time outside the synchronized block.
            try {
                logger.info("About " + remaining + " milliseconds remain for busy workers to quit.");
                Thread.sleep(5000);
                remaining = remaining - DECREMENT;
            } catch (Exception e) { }
        }
        
        // Idle workers should quit fast, give them 10 seconds
        remaining = 10000;
        while (remaining > 0) {
            
            synchronized (targetWorkerCount) {
                if (idleWorkers.size() == 0) break;
                logger.info("Sending quit requests to " + idleWorkers.size() + " idle workers.");
                for (int i=0; i < idleWorkers.size(); i++) idleWorkers.elementAt(i).quit();
            }

            // Spend time outside the synchronized block.
            try {
                logger.info("About " + remaining + " milliseconds remain for idle workers to quit.");
                Thread.sleep(5000);
                remaining = remaining - DECREMENT;
            } catch (Exception e) { }
        }
    }

    
    public boolean assignRequest(I2PSocket socket, long requestId, String fromDestination) {

        synchronized (targetWorkerCount) {
            
            if (shuttingDown) {
                // Job refused, caller must dispose of socket.
                logger.warn("Worker pool shutting down, refusing to assign request " + requestId + "!");
                return false;
            }
            
            if (idleWorkers.isEmpty()) {
                
                try {
                    // Create new worker, hand over references to necessary tools
                    // Register in busy workers pool, start worker, assign socket
                    logger.warn("Assigning request " + requestId + " to new worker, high load?");
                    Worker w = new Worker(this,database,responseFormatter,properties);
                    busyWorkers.addElement(w);
                    w.setName("Worker " + workerId++);
                    w.start();
                    w.assignJob(socket,requestId,fromDestination);
                } catch (Exception e) {
                    // Failed creating worker, caller must dispose of socket.
                    logger.error("Failed creating worker, cannot service connection!",e);
                    return false;
                }
            } else {
                // Pick existing worker, move from idle to busy workers pool, assign socket
                logger.info("Assigning request " + requestId + " to recycled worker.");
                Worker w = idleWorkers.elementAt(0);
                idleWorkers.removeElement(w);
                busyWorkers.addElement(w);
                w.assignJob(socket,requestId,fromDestination);
            }
        }
        // Request assigned, caller may forget about socket.
        return true;
    }

    
    public boolean acceptDone(Worker worker) {

        synchronized(targetWorkerCount) {
        
            String summary = "(" + idleWorkers.size() + " idle, " + targetWorkerCount + " wanted).";

            if (idleWorkers.size() < targetWorkerCount) {
                logger.info("Worker pool accepting " + worker.getName() + " back " + summary);
                idleWorkers.addElement(worker);
                busyWorkers.removeElement(worker);
                return true;
            } else {
                logger.info("Worker pool requesting " + worker.getName() + " to quit " + summary);
                busyWorkers.removeElement(worker);
                return false;
            }
        }
    }
    
    
    public void acceptQuit(Worker worker) {

        synchronized(targetWorkerCount) {
            
            logger.info("Worker pool accepting quit from " + worker.getName() + ".");
            busyWorkers.removeElement(worker);
            idleWorkers.removeElement(worker);
        }
    }

}
